#ifndef _HavranRayCaster_H__
#define _HavranRayCaster_H__

#include "RayCaster.h"
#include "Containers.h"
#include <string>
#include "havran/raypack.h"


namespace GtpVisibilityPreprocessor {


class Intersectable;
class VssRay;
class KdTree;
class Ray;
class RayPacket2x2;
class SimpleRayContainer;
class AxisAlignedBox3;
class Vector3;
struct VssRayContainer;
class Preprocessor;
struct SimpleRay;
class CKTB;
  

/** This class provides an interface for ray casting.
*/
class HavranRayCaster: public RayCaster
{
public:
  /** Default constructor initialising e.g., KD tree
   */
  HavranRayCaster(const Preprocessor &preprocessor);
  virtual ~HavranRayCaster();

  void Build(ObjectContainer &objlist);
  
  int Type() const { return HAVRAN_RAYCASTER; }

  // Using packet of 4 rays supposing that these are coherent
  // We give a box to which each ray is clipped to before the
  // ray shooting is computed !
  virtual void CastRaysPacket4(Vector3 &minBox,
			       Vector3 &maxBox,
			       Vector3 origin4[],
			       Vector3 direction4[],
			       int     result4[],
			       float   dist4[]);
  
  virtual int CastRay(
		      const SimpleRay &simpleRay,
		      VssRayContainer &vssRays,
		      const AxisAlignedBox3 &box,
		      const bool castDoubleRay,
		      const bool pruneInvalidRays = true
		      );
  
  virtual void CastRays16(SimpleRayContainer &rays, 
			  VssRayContainer &vssRays,
			  const AxisAlignedBox3 &sbox,
			  const bool castDoubleRay,
			  const bool pruneInvalidRays = true
			  );

  virtual void CastRays16(SimpleRayContainer &rays,
			  int offset,
			  VssRayContainer &vssRays,
			  const AxisAlignedBox3 &sbox,
			  const bool castDoubleRay,
			  const bool pruneInvalidRays = true
			  );  
  void
  CastSimpleForwardRays(SimpleRayContainer &rays,
			const AxisAlignedBox3 &sbox
			);

  virtual void CastRays(
			SimpleRayContainer &rays,
			VssRayContainer &vssRays,
			const AxisAlignedBox3 &sbox,
			const bool castDoubleRay,
			const bool pruneInvalidRays = true);
protected:
  CKTB *mKtbtree;
  static RayPacket2x2 raypack;
};


} // namespace

#endif
